package Gtk2::Ex::Getopt;
use strict;

use Getopt::Long;
use Gtk2::Ex::Getopt::Base;
use threads;
use IO::Handle;

# these need to be conditional
use Glib qw/TRUE FALSE/;
use Gtk2 '-init';
use Gtk2::Helper;

sub GetOptions(@)
{
    my ($opts, $title, @flags);
    if (ref($_[0]) ne 'HASH')
    {
        die "GetOptions must be passed a hash value\n";
    }

    if (ref @_[1] eq "ARRAY") # no title was passed in
    {
        ($opts, @flags) = @_;
        $title = $0;
    }
    else # title was passed in
    {
        ($opts, $title, @flags) = @_;
    }


    # If the user didn't specify any arguments, we display a GUI for them.
    if ($#main::ARGV == -1 && eval "require Gtk2") # have GTK and have no arguments
    {
        my ($flag, $g, $obj, @objs);
        #print "If we have GTK so we will do something here.\n";

        $g = initGui(\@objs,$opts,$title);

        foreach $flag (@flags)
        {
            $obj = buildGui($flag,$g);
            push @objs, $obj if ($obj);
        }

        runGui($g);

        foreach $obj (@objs)
        {
            # processObj($obj,$opts)
        }
    }
    elsif ($#main::ARGV == -1) # have no arguments but no GTK, so just return
    {
        print "Nothing to do here.\n";
    }
    else # have arguments so skip the UI and pass directly to Getopt::Long
    {
        my (@GetOptflags,$flag);
        foreach $flag (@flags)
        {
            push @GetOptflags, $flag->[0];
        }
        $,=",";
        Getopt::Long::GetOptions($opts,@GetOptflags);
    }
}

sub buildGui
{
    my ($flag, $g) = @_;
    my ($o,$obj,$wid,$label);
    $o = parseOption($flag);

    #print "Type is: ". $o->{'type'} ;
    #print "  Var is: ". $o->{'var'} ;
    #print "  Label is: ". $o->{'label'} . "\n";

    # if label is "!GUI" then don't add the flag to the gui
    if ($o->{'label'} eq "!GUI")
    {
        return undef;
    }


    if ( $o->{'type'} =~ /CheckButton|Entry|Combo|FileOpen|FileSave|SpinButton/ )
    {
        # it's one of the data receiving types
        $obj = "Gtk2::Ex::Getopt::$o->{'type'}"->new($o);
        $wid = $obj->{'gtk'};
        addOption($wid, $o, $g);

        return $obj;
    }
    elsif ($o->{'type'} eq 'Label')
    {
        # Add a blank label to keep spacing right
        $wid = createLabel();
        addOption($wid, $o, $g);
    }
    elsif ($o->{'type'} eq 'FullLabel')
    {
        $wid = createLabel($o->{'label'});
        addTableRow($g,$wid);
    }
    elsif ($o->{'type'} eq 'Separator')
    {
        $wid = Gtk2::HSeparator->new();
        addTableRow($g,$wid);
    }
    elsif ($o->{'type'} eq 'BeginGroup')
    {
        beginGroup($g,$o);
    }
    elsif ($o->{'type'} eq 'EndGroup')
    {
        endGroup($g);
    }
    elsif ($o->{'type'} eq 'Button')
    {
        my ($button,$label);
        $button = Gtk2::Button->new($o->{'label'});
        $wid = Gtk2::HButtonBox->new();
        $wid->add($button);
        $label = createLabel();
        addTableRow($g,$wid,$label);
        if ($o->{'run'})
        {
            $button->signal_connect(clicked => $o->{'run'});
        }
    }
    return undef;
}

# create a label widget out of the $o->{'label'} 
# or out of text if text is passed in
# or blank if nothing is passed in
# justify can be "right"
sub createLabel
{
    my ($o,$justify) = @_;
    my ($label);

    if (ref $o eq "HASH" )
    {
        $label = Gtk2::Label->new($o->{'label'});
    }
    elsif ($o)
    {
        $label = Gtk2::Label->new($o);
    }
    else
    {
        $label = Gtk2::Label->new();
    }
    if ($justify)
    {
        $label->set_justify($justify);       
    }
    return $label;
}

sub addOption
{
    my ($wid, $o, $g) = @_;
    my ($label);
    # if $full is set, then put the widget in across the entire table with no label
    $label = createLabel($o->{'label'},"right");
    if ( $o->{'required'} )
    {
        $label->set_markup("<span foreground='#e00000'>*</span> $o->{'label'}");       
    }
    addTableRow($g,$label,$wid);
}

# if no $right, then center it all the way across
sub addTableRow
{
    my ($g, $left, $right) = @_;

    my ($tr, $h);
    $g->{'tablerows'}++;

    $tr = $g->{'tablerows'};
    $g->{'table'}->resize($tr,2);

    if ($right)
    {
        $g->{'table'}->attach_defaults($right,1,2,$tr-1,$tr);
        $h = Gtk2::HBox->new();
        $h->pack_end($left,0,0,0);
        $g->{'table'}->attach_defaults($h,0,1,$tr-1,$tr);
    }
    else # full widget, so don't add a label
    {
        $g->{'table'}->attach_defaults($left,0,2,$tr-1,$tr);
    }
}

sub parseOption
{
    my ($flag) = @_;
    my ($label,$o);

    my ($var,$format) = parseName (shift @$flag);
    # $format: if type has not been set, CheckButton
    # ! is CheckButton (with --no option), + is SpinButton, = or : is Entry 

    #print "Var: $var  -  format: $format\n";
    $label = shift @$flag;
    if (not $label) # if there is no label then use the var for the label
    {
        $label = $var;
    }
    $o = {@$flag}; # options
    #print "Orig type is " . $o->{'type'} . "\n";
    $o->{'label'} = $label;
    if (not $o->{'type'}) # if the type is not set explicitly, use the type flags from GetOptions
    {
        if ($format =~ /^\+/)    { $o->{'type'} = "SpinButton"; }
        elsif ($format =~ /^[=:]/) { $o->{'type'} = "Entry"; }
        else  { $o->{'type'} = "CheckButton"; }
    }
    if (not $o->{'var'}) # if var is not explicitly set then use the flag as the variable
    {
        $o->{'var'} = $var;
    }

    return $o;
}

sub parseName
{
    my ($spec) = @_;

    $spec =~ m;^
        # whole name pattern
        (?:
            # $1 = Option name
            (\w+[-\w]* )
            # Alias names, or "?"
            (?: \| (?: \? | \w[-\w]* )? )*
        )?
        # $2 = arguments
        (
            # Either modifiers ...
            [!+]
            |
            # ... or a value/dest specification
            [=:] [ionfs] [@%]?
            |
            # ... or an optional-with-default spec
            : (?: -?\d+ | \+ ) [@%]?
        )?
        $;x;

    # map option pieces into variables
    # $optname, $format = ($1, $2);
    return ($1, $2);
}

sub beginGroup
{
    my($g,$o) = @_;
    my($frame,$align);

    # if we already have a group open, close it before 
    if ($g->{'group'})
    {
        endGroup($g);
    }

    # create the frame
    $frame = Gtk2::Frame->new($o->{'label'});
    $frame->set_shadow_type('etched-in');

    # add it to the table
    addTableRow($g,$frame);

    # set group flag
    $g->{'group'} = 1;

    # move old table info to backup
    $g->{'backuptable'} = $g->{'table'};
    $g->{'backuptablerows'} = $g->{'tablerows'};


    # create alignment 
    $align = Gtk2::Alignment->new(0.9, 0.8, 0.9, 0.8);

    # create the new table that goes into the frame
    $g->{'table'} = Gtk2::Table->new(0,2);
    $g->{'table'}->set_col_spacings(8);
    $g->{'tablerows'} = 0;

    # add table to alignment 
    $align->add($g->{'table'});

    # add alignment to the frame
    $frame->add($align);

}
sub endGroup
{
    my($g) = @_;

    $g->{'table'} = $g->{'backuptable'} ;
    $g->{'tablerows'} = $g->{'backuptablerows'};

    $g->{'group'} = 0;

}

sub initGui
{
    my($objs,$opts,$title) = @_;
    my($t,%g,$h);
    # create the window object itself
    $g{'window'} = Gtk2::Window->new('toplevel');
    $g{'window'}->set_title($title);
    $g{'window'}->set_border_width(5);
    $g{'window'}->set_default_size(300,600);
    $g{'window'}->signal_connect(delete_event => \&exitWindow);


    # create a vbox to hold the table and the bottom buttons
    $g{'parentvbox'} = Gtk2::VBox->new();
    $g{'window'}->add($g{'parentvbox'});

    # create a scroll window to hold the table
    $g{'scroll'} = Gtk2::ScrolledWindow->new();
    $g{'scroll'}->set_policy('never','automatic');
    $g{'parentvbox'}->pack_start($g{'scroll'},1,1,0);

    # put the table in the scroll window
    $g{'table'} = Gtk2::Table->new(0,2);
    $g{'table'}->set_col_spacings(8);
    $g{'tablerows'} = 0;
    $g{'scroll'}->add_with_viewport($g{'table'});

    # add hrule
    $h = Gtk2::HSeparator->new();
    $g{'parentvbox'}->pack_start($h,0,0,0);

    # create a cancel button
    $g{'cancel'} = Gtk2::Button->new_from_stock('gtk-cancel');
    $g{'cancel'}->signal_connect(clicked => \&exitWindow);

    # create a OK button
    $g{'ok'} = Gtk2::Button->new_from_stock('gtk-ok');
    $g{'ok'}->signal_connect(clicked => sub { clickOK($objs,$opts,\%g); });
    
    # create an hbuttonbox to put the cancel and go buttons
    $g{'bottom'} = Gtk2::HButtonBox->new();
    $g{'bottom'}->set_layout_default("end");
    $g{'bottom'}->pack_end($g{'cancel'},0,0,0);
    $g{'bottom'}->pack_end($g{'ok'},0,0,0);
    $g{'parentvbox'}->pack_start($g{'bottom'},0,0,0);

    # initialize the group variable to not being in a group
    $g{'group'} = 0;

    #  here's how to add an item:
    #
    #  $testcheck = Gtk2::Ex::Getopt::CheckButton->new();
    #  $g{'testcheckobject'} = $testcheck->gtk;
    #  $$t++;
    #  $g{'table'}->resize($$t,2);
    #  $g{'table'}->attach_defaults($g{'testcheckobject'},1,2,$$t-1,$$t);
    #  $label = Gtk2::Label->new('Click Me!');
    #  $g{'table'}->attach_defaults($label,0,1,$$t-1,$$t);
    
    return \%g;
}

sub clickOK
{
    my ($objs,$opts,$g) = @_;
    my ($obj,$val);

    # first check that all required values exist before processing 
    foreach $obj (@$objs)
    {
        if ($obj->{'required'} && (not $obj->query) )
        {
            my ($dialog) = Gtk2::MessageDialog->new($g->{'window'},
                                      'destroy-with-parent',
                                      'warning', # message type
                                      'close', # which set of buttons?
                                      "Required fields must be set.");
            $dialog->run;
            $dialog->destroy;
            return;
        }
    }

    foreach $obj (@$objs)
    {
        $val = $obj->query;

        if    (ref ($obj->{'var'}) eq 'CODE') 
            { $obj->{'var'}($val); }
        elsif (ref ($obj->{'var'}) eq 'SCALAR')
            { ${$obj->{'var'}} = $val; }
        elsif (ref ($obj->{'var'}) )
            {die;}
        else
            { $opts->{$obj->{'var'}} = $val; }
    }

    Gtk2->main_quit;
}

sub runGui
{

    my ($g) = @_;
    $g->{'window'}->show_all;
    Gtk2->main;

}

sub exitWindow
{
    Gtk2->main_quit;
    exit;
}

# A text window with scroll bars for putting help in
sub textWindow
{
    my ($message) = @_;
    my ($dialog,$buffer,$view,$scroll);
    $dialog = Gtk2::Dialog->new ('Message', undef,
                                    'destroy-with-parent',
                                    'gtk-ok' => 'none');
    $dialog->set_default_size(500,500);
    $buffer = Gtk2::TextBuffer->new();
    $buffer->set_text($message);

    $view = Gtk2::TextView->new_with_buffer($buffer);
    $view->set_editable(0);
    $scroll = Gtk2::ScrolledWindow->new();
    $scroll->set_policy('automatic','automatic');
    $scroll->add ($view);
    $dialog->vbox->add ($scroll);

    # Ensure that the dialog box is destroyed when the user responds.
    $dialog->signal_connect (response => sub { $_[0]->destroy });

    $dialog->show_all;
}

sub make_logwindow
{
    my ($thr, $rdr);
    pipe($rdr, STDOUT);
    open(STDERR, ">&STDOUT");
    autoflush STDOUT 1;
    autoflush STDERR 1;
    $thr = threads->new(\&thread_logwindow);
    close($rdr);
    return;

    END {
        close(STDERR);
        close(STDOUT);
        $thr->join() if ($thr);
    }

    sub thread_logwindow
    {
        my ($buffer) = Gtk2::TextBuffer->new();
        my ($gtk) = Gtk2::Window->new('toplevel');
        $gtk->set_title('Logging');
        $gtk->set_default_size(500,500);
        my ($vb) = Gtk2::VBox->new;
        my ($vw) = Gtk2::TextView->new_with_buffer($buffer);
        $vw->set_editable(0);
        my ($scroll) = Gtk2::ScrolledWindow->new();
        $scroll->set_policy('automatic','automatic');
        $scroll->add ($vw);
        $vb->pack_start($scroll,1,1,0);

        my ($b) = Gtk2::Button->new('Cancel');
        $b->signal_connect(clicked => \&exitlog);

        my ($buttonbox) = Gtk2::HButtonBox->new();
        $buttonbox->set_layout_default("end");
        $buttonbox->pack_end($b,0,0,0);

        $vb->pack_start($buttonbox,0,0,0);
        $gtk->add($vb);
        $gtk->signal_connect(delete_event => \&exitlog);

        $gtk->show_all;
        close(STDOUT);
        close(STDERR);
        my ($watch);
        $watch = Gtk2::Helper->add_watch(fileno($rdr), 'in', sub {
            my ($res) = add_text($buffer, $rdr);
            if ($res eq 'close')
            {
                Gtk2::Helper->remove_watch($watch);
                close($rdr);
                $b->set_label('Close');
            }
            return 1;
        });
        Gtk2->main;
    }

    sub exitlog
    {
#        Gtk2->main_quit;
        exit();
    }

    sub add_text
    {
        my ($buffer, $fh) = @_;
        my ($t, $num);
        return ('close') unless ($num = sysread($fh, $t, 1024));
        return 1 unless ($t);
#        $$ptbuff .= $t;
#        print STDERR "$$ptbuff\n";
#        $buffer->set_text($$ptbuff);
        $buffer->insert($buffer->get_end_iter, $t);
#        print STDERR "$num: " . $buffer->get_text($buffer->get_start_iter, $buffer->get_end_iter, 1);
#        Gtk2->main_quit if (eof($fh));
        return 1;
    }
}

# var => "fred"
# var => \$fred
# var => sub { $fred = $_[0]; }
#
# if (ref $self->{'var'} eq 'CODEREF') 
#   { &$self->{'var'}($val) }
# elsif (ref $self->{'var'} eq 'SCALAR')
#   { &$self->{'var'} = $val; }
# elsif (ref $self->{'var'} )
#   {die;}
# else
#   { $opts{$self->{'var'}} = $val; }

# $name = shift @$row;
# $label = shift @$row
# $hash = {@$row};
# $obj = "Gtk2::Ex::Getopt::$hash->{'type'}"->new($hash);

1;
